Require Import Vtac Program List Arith Even Div2 Relations String.
Require Import basis countable lang subs model basiclemmas weakiso.
Require Import Classical FunctionalExtensionality.

Open Scope string_scope.
Set Implicit Arguments.

Section fiso. (****************************************************************************)

Variables w1 w2 : world.

Structure fiso : Type := 
  { fiso1 : wst w1 -> wst w2
  ; fiso2 : wst w2 -> wst w1
  }.

Definition fiso2wkiso (is : fiso) : iso w1 w2 :=
  mkiso (fun s s' => s' = fiso1 is s) (fun s s' => s' = fiso2 is s).
Coercion fiso2wkiso: fiso >-> iso.

Lemma wexp_fiso : forall (i : fiso) (LI : is_wkiso i) 
    g2 (g2OK: wglob w2 g2) s2 s2' ty e e'
    (WE: wexp w1 (fun s1 => g2 (fiso1 i s1)) (fiso2 i s2) (fiso2 i s2') ty e e'),
  wexp w2 g2 s2 s2' ty e e'.
Proof.
  intros; eapply wexp_iso_aux; eauto.
  replace (morphgl (iso1 i) g2) with (fun s1 => g2 (fiso1 i s1)).
  2 : by extensionality s1; erewrite morphgl_eq; eauto; [apply LI|simpls].
  by eexists; simpls; unfold gc; split; intros; des; subst; eauto.
Qed.

Lemma morphgl_fiso1_eq: forall
    (i: fiso) (LI: is_wkiso i) g2 (g2OK: wglob w2 g2),
  morphgl (iso1 i) g2 = fun s1 => g2 (fiso1 i s1).
Proof.
  by intros; destruct LI; extensionality s2; erewrite morphgl_eq; eauto; simpls.
Qed.

Lemma morphgl_fiso2_eq: forall
    (i: fiso) (LI: is_wkiso i) g1 (g1OK: wglob w1 g1),
  morphgl (iso2 i) g1 = fun s2 => g1 (fiso2 i s2).
Proof.
  by intros; destruct LI; extensionality s2; erewrite morphgl_eq; eauto; simpls.
Qed.

Lemma fiso_glob: forall (i: fiso) (LI: is_wkiso i) g2 (gOK: wglob w2 g2),
  wglob w1 (fun s1 => g2 (fiso1 i s1)).
Proof.
  intros; destruct (id LI); assert (gOK' := gOK).
  eapply morphgl_glob in gOK; eauto.
  erewrite morphgl_fiso1_eq in gOK; eauto.
Qed.

End fiso. (******************************************************************************)


Section swap. (***************************************************************************)

Variables w1 w2 : world_local.

Definition swap_iso  :
  fiso (wlift (wprod w1 w2)) (wlift (wprod w2 w1)) :=
  {| fiso1 := fun s : wst (wlift (wprod w1 w2)) =>
               (fst s, swap (snd s)) : wst (wlift (wprod w2 w1))
   ; fiso2 := fun s => (fst s, swap (snd s))
  |}.

Lemma swap_iso_ok: is_wkiso swap_iso.
Proof.
  intros; econstructor.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro; des; subst; eauto 25.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro; des; subst; eauto 25.
  - red; intros; unfold morph_id in *; destruct M12'; simpls; subst; eauto 10.
  - red; intros; unfold morph_id in *; destruct M12'; simpls; subst; eauto 10.
Qed.

End swap. (******************************************************************************)


Section id. (****************************************************************************)

Variables w : world_local.

Definition id_intro_iso : fiso (wlift (wprod w wid)) (wlift w) :=
  {| fiso1 := fun s : wst (wlift (wprod w wid)) => (fst s, fst(snd s)) : wst (wlift w)
   ; fiso2 := fun s => (fst s, (snd s,()))
  |}.

Ltac elim_hemp :=
  repeat match goal with 
  | [H: hplus ?h hemp ?h' |- _] => apply hplus_hemp_rE in H; subst h' 
  | [H: hplus hemp ?h ?h' |- _] => apply hplus_hemp_lE in H; subst h'
  end. 

Lemma id_intro_ok: is_wkiso id_intro_iso.
Proof.
  intros; econstructor.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 25; []; destruct ty; contradiction.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 25; []; destruct ty; contradiction.
  - red; intros; unfold morph_id in *; destruct M12'; simpls; subst; eauto 10.
  - red; intros; unfold morph_id in *; destruct M12'; simpls; subst; eauto 10.
Qed.

Definition id_elim_iso : fiso wref (wlift wid) :=
  {| fiso1 := fun s : wst wref => (s,()) : wst (wlift wid)
   ; fiso2 := fun s => fst s
  |}.

Lemma id_elim_ok: is_wkiso id_elim_iso.
Proof.
  intros; econstructor.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 20; []; destruct ty; contradiction.
  - econstructor; simpl; unfold nunion; repeat split; repeat intro;
    simpls; des; subst; elim_hemp; eauto 20; []; destruct ty; contradiction.
  - red; intros; unfold morph_id in *; destruct M12'; simpls; subst; eauto 10.
  - red; intros; unfold morph_id in *; destruct M12'; simpls; subst; eauto 10.
Qed.

Lemma wglob_wid: forall g (gOK : wglob (wlift w) g) s0,
  wglob (wlift wid) (fun s => g (fst s,s0)).
Proof.
  intros; eapply (fiso_glob (id_intro_ok)) in gOK.
  by eapply wglob_prod2 in gOK; simpls; eauto.
Qed.

End id. (********************************************************************************)

