Require Import Vtac Program List Arith Even Div2 Relations String.
Require Import basis lang subs model basiclemmas weakiso.

Section WeakIsoExamples.

(*----------------------------------------------------------------------------*)
(** * Erased World Isomorphism *)
(*----------------------------------------------------------------------------*)

Program Definition werased (w: world) : world :=
  {| wnam := wnam w
   ; wst := {s : wst w | exists g h1 h2, wglob w g /\ wheap w () s (g s) h1 h2}
   ; wext s' s := wext w (`s') (`s)
   ; wextpub s' s := wextpub w (`s') (`s)
   ; wval ps s := wval w ps (`s)
   ; wheap ps s := wheap w ps (`s) |}.
Next Obligation. eauto using wext_trans. Qed.
Next Obligation. eauto using wextpub_trans. Qed.
Next Obligation. eauto using wval_mon. Qed.
Next Obligation. eapply wval_closed; eauto. Qed.
Next Obligation. eauto 10 using wheap_mon. Qed.
Next Obligation. eauto using wnam_supports. Qed.
Next Obligation. eauto using wval_arrow_wf. Qed.
Next Obligation. eauto using wval_forall_wf. Qed.

Program Definition erasediso w : iso w (werased w) :=
  {| iso1 s' s := s' = `s
   ; iso2 s s' := s' = `s |}.

Lemma erasediso_is_wkiso: forall w, is_wkiso (erasediso w).
Proof.
  constructor.
  - constructor; repeat intro; simpls; subst; eauto; intuition.
    assert (pf: exists g h3 h4, wglob w g /\ wheap w () s1 (g s1) h3 h4) by eauto.
    by exists (exist _ s1 pf : wst (werased w)); eauto.
  - by constructor; repeat intro; simpls; subst; eauto; intuition.
  - repeat intro; simpls; unfold morph_id in *; subst.
    by destruct M12'; subst; rewrite M23; eauto.
  - repeat intro; simpls; unfold morph_id in *; subst.
    by destruct M12'; subst; eauto.
Qed.

Lemma weq_erase: forall D G w ty e e',
  weq D G w ty e e' <-> weq D G (werased w) ty e e'.
Proof.
  by red; eauto using weq_iso, is_wkiso_sym, erasediso_is_wkiso.
Qed.

(*----------------------------------------------------------------------------*)
(** * Flat World Isomorphism *)
(*----------------------------------------------------------------------------*)

Program Definition wflat (w: world) : world :=
  {| wnam := wnam w
   ; wst := wst w * (heap * heap)
   ; wext s' s := wext w (fst s') (fst s)
   ; wextpub s' s := wextpub w (fst s') (fst s)
   ; wheap ps s r h1 h2 := snd s = (h1,h2) /\ wheap w ps (fst s) r h1 h2
   ; wval ps s := wval w ps (fst s) |}.
Next Obligation. eauto using wext_trans. Qed.
Next Obligation. eauto using wextpub_trans. Qed.
Next Obligation. eauto using wval_mon. Qed.
Next Obligation. eapply wval_closed; eauto. Qed.
Next Obligation. des; eauto 10 using wheap_mon. Qed.
Next Obligation. eauto using wnam_supports. Qed.
Next Obligation. eauto using wval_arrow_wf. Qed.
Next Obligation. eauto using wval_forall_wf. Qed.

Program Definition flatiso w : iso w (wflat w) :=
  {| iso1 s' s := s' = fst s
   ; iso2 s s' := s' = fst s |}.

Lemma flatiso_is_wkiso: forall w, is_wkiso (flatiso w).
Proof.
  constructor.
  - constructor; repeat intro; simpls; subst; eauto; intuition.
    by exists (s1,(h1,h2)); eauto.
  - constructor; repeat intro; simpls; subst; eauto; intuition.
    by exists (fst s1); eauto.
  - repeat intro; simpls; unfold morph_id in *; subst.
    by destruct M12'; subst; rewrite M23; eauto.
  - repeat intro; simpls; unfold morph_id in *; subst.
    by destruct M12'; subst; eauto.
Qed.

Lemma weq_flat: forall D G w ty e e',
  weq D G w ty e e' <-> weq D G (wflat w) ty e e'.
Proof.
  by red; eauto using weq_iso, is_wkiso_sym, flatiso_is_wkiso.
Qed.

End WeakIsoExamples.

